///////////////////////////////////////////////////////////////////////////////
//
// Description
//
//  A simple string class. Strings can be of any (i.e. unlimited) length up
//  to the limit of 'unsigned' on a machine.
//
// Edit History
//
//      Written October 1993 by Andrew Davison.
//
///////////////////////////////////////////////////////////////////////////////

#ifndef ASTRING_H
#define ASTRING_H

#include <string.h>

class istream;
class ostream;

class AString
{
private:

	static AString nil;

	struct AStringRef
	{
		unsigned refcnt;
		unsigned nbytes;
		char* buf;
	}
	  *sref;

protected:

	AString& Assign(const AString& s) { return *this = s; };

public:

	AString() { sref = 0; };
	AString(char);
	AString(const char*);
	AString(const char*, unsigned len);
	AString(const AString&, unsigned offset, unsigned len);
	AString(const AString& s);
	~AString();

	unsigned size() const { return (unsigned)(sref ? sref->nbytes : 0); };
	unsigned length() const { return size(); };
	int empty() const { return size() == 0; };

	// Assignment...

	AString& operator=(char);
	AString& operator=(const char*);
	AString& operator=(const AString&);

	// Concatenation...

	AString& operator+=(char);
	AString& operator+=(const char*);
	AString& operator+=(const AString&);

	// Addition...

	AString operator+(char) const;
	AString operator+(const char*) const;
	AString operator+(const AString&) const;

	// Equality...

	int operator==(char* s) const;
	int operator==(const char*) const;
	int operator==(const AString&) const;

	int operator!=(char* s) const;
	int operator!=(const char*) const;
	int operator!=(const AString&) const;

	// Maintenance (note: these operate on the string itself)...

	AString& pad(unsigned len, char ch = ' ');
	AString& trim(char ch = ' ');
	AString& upper();
	AString& lower();

	// Sub-string operations...

	AString left(unsigned len) const;
	AString right(unsigned len) const;
	AString mid(unsigned offset, unsigned len) const;
	AString dq() const;
	AString pack() const;
	AString unpack() const;
	AString change(char from, char to) const;
	AString replace(const char* from, const char* to) const;

	// Cast...

	const char* c_str() const { return (sref ? sref->buf : nil.c_str()); };
	const char* data() const { return c_str(); };

	// Access...

	char operator [] (int n) const
		{ return (char) (sref ? ((unsigned)n < sref->nbytes ? sref->buf[n] : 0) : 0); };

	char operator [] (unsigned n) const
		{ return (char) (sref ? ((unsigned)n < sref->nbytes ? sref->buf[n] : 0) : 0); };

	// Output...

	friend ostream& operator<<(ostream&, const AString&);
	friend istream& operator>>(istream&, AString&);

	// Concatenation...

	friend AString operator+(char, const AString&);
	friend AString operator+(const char*, const AString&);

	// WXWINDOWS compatability...

	const char* Chars() const { return c_str(); };
	unsigned Length() const { return size(); };
	AString At(unsigned offset, unsigned len) const { return mid(offset, len); };
	char Lastchar() const { return (size()?(*this)[size()-1]:0); };
	int Contains(const char* str) const;
};

inline AString::AString(const AString& s)
{
	sref = s.sref;
	if (sref) sref->refcnt++;
}

inline AString::~AString()
{
	if (sref)
	{
		if (--sref->refcnt == 0)
		{
			delete [] sref->buf;
			delete sref;
		}

		sref = 0;
	}
}

inline int AString::operator==(const char* s) const
{
	return sref && !strcmp(sref->buf, s);
}

inline int AString::operator==(char* s) const
{
	return *this == (const char*)s;
}

inline int AString::operator==(const AString& s) const
{
	if (sref == s.sref)
		return 1;
	else if (!sref || !s.sref)
		return 0;
	else if (sref->nbytes != s.sref->nbytes)
		return 0;
	else
		return !strcmp(sref->buf, s.sref->buf);
}

inline int AString::operator!=(const char* s) const
{
	return !sref || strcmp(sref->buf, s);
}

inline int AString::operator!=(char* s) const
{
	return *this != (const char*)s;
}

inline int AString::operator!=(const AString& s) const
{
	if (sref == s.sref)
		return 0;
	else if (!sref || !s.sref)
		return 1;
	else if (sref->nbytes != s.sref->nbytes)
		return 1;
	else
		return strcmp(sref->buf, s.sref->buf);
}

#endif
